/*  Copyright (c) 2016-2025 Daniel Kos, General Development Systems

    This file is part of Hipe.

    Hipe is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hipe is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hipe.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef CLIENTCONTAINER_H
#define CLIENTCONTAINER_H

#include <QtWebKit/QWebElement>
#include <QtWebKitWidgets/QWebFrame>
#include <QObject>
#include <stack>
#include <list>
#include <map>
#include <string>
#include "ExpArray.hh"
#include "common.h"
#include "keylist.h"
class Connection;

#define MAX_N_STYLESHEETS 20
//the number of elements allocated to the the stylesheets array, 
//where each element index corresponds to a different CSS theme.


class FrameData {
//stores metadata and references relevant to a particular child container.
public:
    QWebElement we; //reference to the <iframe> tag
    QWebFrame* wf; //reference to the frame object which matches the <iframe> tag.
    QString hostkey; //once claimed, this becomes an empty string.

    uint64_t requestor; //to be attached when sending back child frame events.
    std::string clientName;
    std::string title;
    pid_t pid; //client process ID. pid_t defined in <types.h>

    std::string fg, bg; //foreground and background colours as CSS values.

    bool operator==(const FrameData& other);
};


class Container : public QObject
{
    Q_OBJECT
public:
    friend class ContainerManager;
    static std::string globalStyleSheets[MAX_N_STYLESHEETS];
    static int maxLoadedStyleSheetIndex;


    Container(Connection* client, std::string clientName, int themeIndex);
    ~Container();

    void receiveInstruction(hipe_instruction instruction);

    virtual void setTitle(std::string newTitle)=0;
    virtual void setIcon(const char* imgData, size_t length)=0;
    virtual void setBody(std::string newBodyHtml, bool overwrite=true)=0;

    QColor fg, bg,  cursorColor1, cursorColor2, cursorColor3;
    //stores the current foreground and background colors, and derivative colors
    // for drawing secondary details like embedded mouse cursor.

    void applyStylesheet();
    //apply stylesheet after changes. If <body> was not opened yet (!initYet)
    //then this is a no-op as styling gets applied when setBody is called.
    //Otherwise, this call causes a <style> tag to be appended inside the
    //<body> tag; which is not technically valid HTML but should do the trick.

    void containerClosed();

    Container* requestNew(std::string key, std::string clientName, uint64_t pid,
                        int themeIndex, Connection* c);
    //request a new sub-frame that is managed by this container. Returns nullptr if the
    //key is not held by this container, else creates a new ContainerFrame object and
    //returns its pointer.

    void receiveSubFrameEvent(short evtType, QWebFrame* sender, std::string detail);
    //called by a sub-frame (ContainerFrame object) of this container to indicate that
    //the frame has been modified in a way that should be reported to the framing client.

    void receiveMessage(char opcode, int64_t requestor, const std::vector<std::string>& args, QWebFrame* sender, bool propagateToParent=false);
    //called by another container object to transmit an instruction (e.g. HIPE_OP_MESSAGE)
    //from a direct parent/child frame's client to this container's client.
    //sender should be used only if sender is a child of the recipient. If it's
    //the parent, this should be indicated by passing a nullptr.
    //This can also be used to transmit events across frame boundaries
    //(the parent will see the event as originating from the client frame element).

    void keyEventOnChildFrame(QWebFrame* origin, bool keyUp, QString keycode);
    //if keyup is false, it was a keydown event.
    //This function is called from a child container instructing this container
    //that a keyup/keydown event has occurred on the body element of this frame
    //(or has propagated from a child frame of *that* frame).
    //The event should be propagated up to the top level so the framing manager
    //can intercept global keyboard shortcuts.
    //It should also trigger a simulated event on the frame to this client,
    //if this client has bound onkeydown/onkeyup attributes to this frame.

    virtual Container* getParent()=0; //returns the parent container, or nullptr if it's a top level container.

    virtual char editActionStatus(char) {return '?';}

    bool bgColorChanged(std::string newColor);
    bool fgColorChanged(std::string newColor);
    //call to update the container on its new fg or bg color.
    //The container will store the new color and update the embedded cursor if applicable
    //Returns true if the new color is different to the old color, otherwise
    //returns false if the color has not changed.


    QWebFrame* frame;

    Connection* client;
    QWebElement webElement;
    std::string stylesheet; //build up the stylesheet before we apply it.

    bool isTopLevel = false; //some instructions are only permitted to be carried out
    //by the top level frame.
protected:
    bool initYet; //becomes true once boilerplate html has been set.

#ifndef HAVE_HIPECORE  //without hipecore, need to handle events via javascript
    signals:
    void receiveGuiEvent(QString location, QString requestor, QString event, QString detail);
    //signal called from within the QWebView object (via Javascript), each time a user interaction takes place.

    void receiveKeyEventOnBody(bool keyUp, QString keycode);
    //signal called when a keyup (or else keydown) event happens on the body element.
protected slots:
    void _receiveGuiEvent(QString location, QString requestor, QString event, QString detail);
    //location and requestor are to be passed as hexadecimal strings.

    void _receiveKeyEventOnBody(bool keyUp, QString keycode);
#else  //HAVE_HIPECORE
    static void _receiveKeyEventOnBody(const QString& eventName, void* containerPtr, uint64_t isKeyUp,
                                        uint64_t requestor, const QString& eventDetails);

    static void _receiveDragStartEvent(const QString&, void*, uint64_t,
                                        uint64_t, const QString&) {};
#endif //HAVE_HIPECORE
protected slots:
    void frameCleared();
    void frameDestroyed(); //conneected to the QWebFrame's destroyed() signal.



protected:
private:
    ///This block of variables/functions provides pointer protection of web
    ///element references so a handle (array element number) can safely be given
    ///to external processes, and invalid references can be detected.
    AutoExpArray<QWebElement*> referenceableElement;
    size_t firstFreeElementAfter=1;
    //store the location of the first free element, or a smaller element number, to speed insertions.
    size_t maxElementIndexUsed=0;
    //store the maximum location index used so far in this container to disallow
    //growing the index too quickly.
public:
    //when an element is created, we store its location as an element number here, then share the element number
    //(not a QWebElement--dangerous!) with the client.
    size_t assignElementIndex(const QWebElement& w, size_t newIndex);
    void removeReferenceableElement(size_t);
    QWebElement getReferenceableElement(size_t); //resolve a reference integer.
    size_t findReferenceableElement(const QWebElement&);
    size_t getIndexOfElement(const QWebElement&);
    //finds corresponding index, or adds it if it has not been allocated an index yet.

    //flags to handle keyup/down events on body as a special case (since this event needs to propagate to the framing manager for special window manipulation keys)
    bool reportKeydownOnBody=false; //has the client requested keydown events on the body element?
    bool reportKeyupOnBody=false; //has the client requested keyup events on the body element?
    uint64_t keyDownOnBodyRequestor=0; //corresponding requestors if the client requests these key events.
    uint64_t keyUpOnBodyRequestor=0;

public:
    KeyList* keyList;
    std::list<FrameData> subFrames;
    //table of subframes, mapping web element of an iframe to its corresponding
    //child frame object and host-key (if assigned).


};

#endif // CLIENTCONTAINER_H
