/*
    Copyright (C) 2025 Daniel Kos

    This file is part of Hipe.

    Hipe is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Hipe is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Hipe.  If not, see <http://www.gnu.org/licenses/>.
*/


//implements a mouse cursor sprite for EGL systems where a hardware cursor is unavailable.
//May not behave correctly within an existing desktop environment.


#include "mousecursor.h"
#include "container.h"


#include <QApplication>
#include <QScreen>
#include <QPixmap>

MouseCursor::MouseCursor(Container* c, QWidget* parent) : QWidget(parent) {
    this->c = c;

    // Make the widget transparent and frameless
    setAttribute(Qt::WA_TranslucentBackground);
    setAttribute(Qt::WA_TransparentForMouseEvents);
    setWindowFlags(Qt::FramelessWindowHint | Qt::WindowStaysOnTopHint);
    


    QScreen* screen = QGuiApplication::primaryScreen();
    if (screen) {
        resize(screen->size());
    }
    
    // Start timer for cursor updates
    startTimer(16);  // ~60 FPS

    // Hide the system cursor
    QGuiApplication::setOverrideCursor(Qt::BlankCursor);

    qApp->installEventFilter(this);
}

MouseCursor::~MouseCursor() {
    qApp->removeEventFilter(this);
}

bool MouseCursor::eventFilter(QObject* obj, QEvent* event) {
    switch (event->type()) {
        case QEvent::MouseButtonPress:
            mousePressed = true;
            update();
            break;
        case QEvent::MouseButtonRelease:
            mousePressed = false;
            update();
            break;
        default:
            break;
    }
    return QWidget::eventFilter(obj, event);
}

void MouseCursor::paintEvent(QPaintEvent*) {

    QPainter painter(this);
    //painter.setRenderHint(QPainter::Antialiasing, true);  // un-Disable antialiasing
    
    QPoint pos = mapFromGlobal(QCursor::pos());
    
    // Draw white crosshair lines
    //painter.setPen(QPen(QColor(0,255,0,127), 1));
    painter.setPen(QPen(!mousePressed ? c->cursorColor1 : c->cursorColor3, 1));
    painter.drawLine(pos.x(), 0, pos.x(), height());
    painter.drawLine(0, pos.y(), width(), pos.y());
    
    // Draw centre circle
    if(!mousePressed) {
        painter.setPen(QPen(c->cursorColor2,3));
        painter.drawEllipse(pos.x() - 8, pos.y() - 8, 16, 16);
        painter.setPen(QPen(c->cursorColor3,1));
        painter.drawEllipse(pos.x() - 8, pos.y() - 8, 16, 16);
    } else {
        painter.setPen(QPen(c->cursorColor2,3));
        painter.drawEllipse(pos.x() - 6, pos.y() - 6, 12, 12);
        painter.setPen(QPen(c->cursorColor3,1));
        painter.drawEllipse(pos.x() - 6, pos.y() - 6, 12, 12);
    }
}

void MouseCursor::timerEvent(QTimerEvent*) {

    static QPoint lastPos = QCursor::pos();
    QPoint currentPos = QCursor::pos();
    
    if (currentPos != lastPos) {
        update();
        lastPos = currentPos;
    }
}


